#!/usr/bin/php -q
<?php
/*
 * IPSTools for LINUX
 *
 * Dies ist eine PHP-basierte Version der IPSTools für Linux.
 *
 * Author:  Thorsten Bruns  <thorsten.bruns@gmx.de>
 * Version: V0.2alpha
 *
 * Version      Datum       Name         Beschreibung
 * """"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
 * V0.1alpha    23.12.2013  Bruns        Erste lauffähige Version, entlassen in die Wildniss
 * V0.2alpha    24.12.2013  Bruns        Diverse kosmetische Code-updates
 *                                       Mouse-Idle-Überwachung eingebaut
 *                                       Parameterformat angepasst. Jetzt werden Parameterwerte per ':' angehängt
 *                                       Kleines C-Programm zur Abfrage der Mausposition hinzugefügt
 *                                       Parameterbeschreibung für die Hilfe angepasst
 * V0.3alpha    30.12.2013  Bruns        Zugriffspfad für Tool getMousePos angepasst, da './' nicht zuverlässig funktioniert.
 */

// Grundlegende PHP-Einstellungen anpassen
set_time_limit(0);
ob_implicit_flush ();

// Zulässige Parameter und ihre Default-Werte
$options = array(
    'help' => false,
    'bind' => '0.0.0.0',
    'port' => 7002,
    'send-idle' => false,
    'send-delay' => 5,
    'shutdown-pin' => '1234',
    'time-zone' => 'Europe/Berlin',
    'debug' => false,
    'jitter' => 5,
);

// Kommandozeilenparameter verarbeiten
$argc = count($argv);
for($i = 1; $i < $argc; $i++) {
  $arg = $argv[$i];
  printf("Handling parameter No. %d: <%s>\n", $i, $arg);
  $opt = null;
  if (strpos($arg, ':') !== false) {
    echo "\tSplitting up value into optionsname and optionvalue\n";
    list($arg, $opt) = explode(':', $arg);
  }
  if (substr($arg, 0, 2) == '--' && isset($options[substr($arg, 2)])) {
    if (is_null($opt)) {
      printf("Setting Parameter: %14s => <%s>\n", '['.substr($arg, 2).']', 'TRUE'); 
      $options[substr($arg, 2)] = true;
    } else {
      printf("Setting Parameter: %14s => <%s>\n", '['.substr($arg, 2).']', $opt); 
      $options[substr($arg, 2)] = $opt;
      $i++;
    }
  }
}

// Das PHP-Error-Reporting einstellen (E_ALL wenn debug, sonst 0)
if ($options['debug']) echo 'setting php error reporting to '.($options['debug'] ? 'E_ALL' : 'NONE')."\n";
if ($options['debug']) {
  error_reporting(E_ALL);
} else {
  error_reporting(0);
}

// Default-Zeitzone einstellen
if ($options['debug']) echo 'Setting default timezone to '.$options['time-zone']."\n";
date_default_timezone_set($options['time-zone']);

// Hilfemodus. Zeigt erlaubte Parameter und beendet sich dann gleich wieder
if ($options['help'] == true) {
  echo GetHelpScreen();
  die();
}


// Socket-Instanz erstellen
if (($sock = socket_create(AF_INET, SOCK_STREAM, SOL_TCP)) === false) {
  echo "socket_create() fehlgeschlagen: Grund: " . socket_strerror(socket_last_error()) . "\n";
  die();
}
if (socket_bind($sock, $options['bind'], $options['port']) === false) {
  echo "socket_bind() fehlgeschlagen: Grund: " . socket_strerror(socket_last_error($sock)) . "\n";
  die();
}
if (socket_listen($sock, 0) === false) {
  echo "socket_listen() fehlgeschlagen: Grund: " . socket_strerror(socket_last_error($sock)) . "\n";
  die();
}
if ($options['debug']) echo "\nServer ist bereit!\n";
// Haupt-Server-Schleife
do {
  if (($msgsock = socket_accept($sock)) === false) {
    echo "socket_accept() fehlgeschlagen: Grund: " . socket_strerror(socket_last_error($sock)) . "\n";
    break;
  }
  if ($options['debug']) echo "\nClient connected!\n";

  // Client-Schleife vorbereiten

  // Aktuelle Mausposition holen und in $mousePos ablegen
  $mousePos   = shell_exec(__DIR__.'/getMousePos');

  // Sende- und Idle-Timer mit der aktuellen Systemzeit initialisieren
  $sendTimer  = time();
  $mouseTimer = time();

  // Empfangspuffer vorbereiten
  $buffer = '';

  if ($options['debug']) echo "\tStarting client loop\n";
  do {

    // Daten vom Socket in einer nicht-blockierenden Schleife abholen - damit kann nebenbei noch die Mouse-Idle-Time erfasst werden
    $wait = true;
    do {

      // Mouse-Idle-Time aktualisieren
      $idleTime = checkMouseIdleState($mousePos, $mouseTimer);

      // Den aktuellen Idle-Wert senden, falls aktiviert und delay abgelaufen
      if ($options['send-idle'] && (time() > ($sendTimer + $options['send-delay']))) {
        if ($options['debug']) echo "\tSend client update of mouse idle time\n";
        $msg = 'MouseIdleSince;'.$idleTime."\n";
        socket_write($msgsock, $msg, strlen($msg));
        $sendTimer = time();
      }

      // Prüfen ob neue Daten am Socket angekommen sind...
      $r = $w = $e = NULL;
      $r = array($msgsock);
      if (socket_select($r, $w, $e, 0) < 1) {
        // Keine neuen Daten, 1 Sekunde pause, dann wieder von vorne anfangen
        sleep(1);
        continue;
      }

      // Bis zu 512 Bytes vom Socket holen
      $data = @socket_read($msgsock, 512, PHP_NORMAL_READ);

      // Wenn der client inzwischen die Verbindung beendet hat haben wir false in $data
      if ($data === false) {
        break 2;
      }

      // Die gelesenen Daten an den Buffer anhängen und auf vollständigen Request prüfen
      $buffer .= $data;
      if (strpos($buffer, "\n") !== false) {
          list($request, $buffer) = explode("\n", $buffer);
          $wait = false;
          $request = trim($request);
      }
    } while ($wait);
    if ($options['debug']) echo "\tRequest received: [$request]\n";

    if ($request == 'quit') {
      break;
    }
    if (substr($request, 0, 8) == 'shutdown') {
      $code = trim(substr($request, 8));
      if ($code == $options['shutdown-pin']) {
	socket_close ($msgsock);
	break 2;
      } else {
        // Falscher PIN? Dann schmeissen wir einfach den Client raus und gut is...
        break;
      }
    }

    $response = handle_request($request)."\n";
    socket_write ($msgsock, $response, strlen ($response));
  } while (true);
  socket_close ($msgsock);
  if ($options['debug']) echo "\nClient disconnected!\n";
} while (true);

if ($options['debug']) echo "\nStopping Server\n";
socket_close ($sock);


function handle_request($raw_request) {
  global $options;

  // Den Request zerlegen in Request und ggf. Parameter. Alles muss mit ; getrennt sein
  $request = explode(';', $raw_request);
  $response = '';

  // In $request[0] steht jetzt der Requestname, alle folgenden sind ggf. Parameter.
  switch($request[0]) {
  case 'ScreenPowerOff':
    SetScreenPower(false);
    $response = 'ScreenPowerOff';
    break;

  case 'ScreenPowerOn':
    SetScreenPower(true);
    $response = 'ScreenPowerOn';
    break;

  case 'MousePosition':
    $mousePos = GetMousePosition();
    list($x, $y) = explode('|', $mousePos);
    $response = "MousePosition;$x;$y";
    break;

  case 'MousePositionRaw':
    $mousePos = GetMousePosition();
    $response = "MousePositionRaw;$mousePos";
    break;
    
  case 'SystemUpTime':
    $response = 'SystemUpTime;'.GetSystemUptime();
    break;

  case 'AvailableMemory':
    $response = 'AvailableMemory;'.GetAvailableMemory();
    break;

  case 'UsedCPU':
    $response = 'UsedCPU;'.GetLoadAverage();
    break;

  case 'commands':
  case 'help':
    $response = get_help_screen();
    break;

  default:
    if ($options['debug']) echo "\nUnbekannter Request: $raw_request\n";
    $response = "Unbekannter Request: $raw_request";
    break;
  }

  return $response;
}

function get_help_screen() {
  return
    "\n\n".
    'Verwendung: IPSToolsForLinux [OPTIONEN]...'."\n".
    "\n".
    'Verfügbare OPTIONEN:'."\n".
    '--help                    Zeigt die Parameterhilfe an.'."\n".
    '--bind:<adresse>          Legt das Interface fest, auf das sich der Server bindet. Default ist 0.0.0.0,'."\n".
    '                          welches allen verfügbaren Interfaces entspricht.'."\n".
    '--port:<portnummer>       Left die Portnummer fest, auf die der Server bindet. Default ist 7002.'."\n".
    '--send-idle               Wird dieser Parameter gesetzt, wird das zyklische Senden des Maus-Idle-Timers'."\n".
    '                          aktiviert. Default ist inaktiv, d.h. der client muss die Zeit per MouseIdleSince'."\n".
    '                          gezielt abfragen.'."\n".
    '--send-delay:<delay>      Legt die Zykluszeit (in Sekunden) für das automatische Senden des'."\n".
    '                          Maus-Idle-Timers fest. Default ist 1s.'."\n".
    '--use-timezone:<zoneID>   Legt die zu verwendene Zeitzone fest. Default ist <Europe/Berlin>'."\n".
    "\n".
    'Folgende Requests werden vom Server verarbeitet:'."\n".
    ' ScreenPowerOff            Schaltet den Monitor aus (xset dpms force off).'."\n".
    ' ScreenPowerOn             Schaltet den Monitor ein (xset dpms force on).'."\n".
    ' StartScreenSaver          Startet den Bildschirmschoner der X-Session.'."\n".
    '                           (noch nicht implementiert)'."\n".
    ' MousePosition             Liefert die aktuelle Mausposition zurück. XPos=Abs. horz. Pos. in Pixel, YPos=abs. vert. Pos. in Pixel'."\n".
    ' MouseIdleSince            Liefert die Zeit seit der letzten Mausbewegung. Sec=Anzahl Sekunden seit letzter Bewegung.'."\n".
    ' FreeDiscSpace;<Drive>     Wird unter Linux nicht unterstützt.'."\n".
    ' TotalDiscSpace;<Drive>    Wird unter Linux nicht unterstützt.'."\n".
    ' DriveType;<Drive>         Wird unter Linux nicht unterstützt.'."\n".
    ' DriveFormat;<Drive>       Wird unter Linux nicht unterstützt.'."\n".
    ' IsDriveAvailable;<Drive>  Wird unter Linux nicht unterstützt.'."\n".
    ' SystemUpTime              Liefert die Zeit seit dem letzten Systemstart. <Sec>=Laufzeit in Sekunden.'."\n".
    ' AvailableMemory           Liefert den freien Systemspeicher des Systems. <Value>=Freier Speicher in kB.'."\n".
    ' UsedCPU                   Liefert die Systemauslastung des Systems. <Value>=Auslastung in Prozent 0-100.'."\n".
    "\n\n";
}

function SetScreenPower($status) {

  if ($status) {
    $result = shell_exec('xset dpms force on');
  } else {
    $result = shell_exec('xset dpms force off');
  }

  return $result;
}

function GetSystemUptime() {

    // Read /proc/uptime
    $uptime = file_get_contents('/proc/uptime');

    // Separate uptime and idle value
    $list = explode(' ', $uptime);
    $up   = (float)$list[0];

    // Return uptime value
    return $up;
}

function GetAvailableMemory() {

  // Read /proc/meminfo
  $meminfo = file_get_contents('/proc/meminfo');

  // Extract informations:

  // Step 1: Split content into single values
  $list = explode("\n", $meminfo);
  // Step 2: Walk thru the list and copy keys and values to info
  $info = [];
  foreach($list as $line) {
    // Split Dataline up into Key and Value
    $parts = explode(':', $line);
    // Now trim key and value
    $key   = trim($parts[0]);
    $value = trim($parts[1]);

    // Check if value have a unit string at the end, therefor split value at space char
    $parts = explode(' ', $value);
    // Now we can first convert value into an float
    $value = (float)$parts[0];
    // If we have an unit it is in $parts[1]
    if (isset($parts[1])) {
      // We have... identify and recalculate value accordingly
      switch (strtolower($parts[1])) {
      case 'kb':
        $value *= 1024;
        break;
      case 'mb':
        $value *= 1048576;
        break;
      case 'gb':
        $value *= 1073741824;
        break;
      }
    }
    // Now save result in $info
    $info[$key] = $value;
  }
  // Now calculate the free space, which is MemFree + Buffers
  $free = $info['MemFree'] + $info['Buffers'];

  return $free;
}

function GetCpuCount() {

  // Read /proc/cpuinfo
  $cpuinfo = file_get_contents('/proc/stat');

  // Use regex search to find all /CPU[0-9] / and count number of hits, this is equal to the number of cpu's available
  preg_match_all("/cpu[0-9] /", $cpuinfo, $result);

  var_dump($result);

  return count($result);
}

function GetLoadAverage() {

  // Read /proc/loadavg
  $loadavg = file_get_contents('/proc/loadavg');

  // Split up into values for avg 1m, avg 5m, avg 15m, scheduler-count and pid of last created process
  $parts = explode(' ', $loadavg);

  // Get load average for 1 minute cycle
  $load = (float)$parts[0];

  // Scale to load in percent. Therefor first get number of cpu's (and convert to float)
  $cpu = 0.0 + GetCpuCount();

  // Now calculate load value as:  load = ($load / $cpu) * 100.0
  $result = ($load / $cpu) * 100.0;

  return $result;
}

function GetMousePosition() {

  // Aktuelle Mausposition holen
  return shell_exec(__DIR__.'/getMousePos');
}  
  
function CheckMouseIdleState(&$position, &$idleTimer) {
  global $options;

  // Aktuelle Mausposition holen
  $mousePos = GetMousePosition();

  // Beide Positionswerte zerlegen in X- und Y-Position
  list($lastX, $lastY) = explode('|', $position);
  list($actX, $actY)   = explode('|', $mousePos);

  // Die Differenzwerte für X und Y bilden
  $diffX = abs($lastX - $actX);
  $diffY = abs($lastY - $actY);
  if ($options['debug']) echo sprintf("\tLast Mousepos: %d|%d\tAct. Mousepos: %d|%d\t  Diff: %d|%d\tActual Timer: %d\tActual time: %d\n", 
                                      $lastX, $lastY, $actX, $actY, $diffX, $diffY, $idleTimer, time());

  // Prüfen ob eine der Achsen sich um mehr als 5 Pixel verändert hat
  if(($diffX > $options['jitter']) || ($diffY > $options['jitter'])) {
    // Maus wurde bewegt, idleTimer neu starten
    if ($options['debug']) echo "\tMausbewegung erkannt. Tiemr wird neu gestartet...\n";
    $idleTimer = time();
  }

  // Neue Position sichern
  $position = $mousePos;

  return (time() - $idleTimer);
}
